import os
import json
from pathlib import Path
import pytest
from mock import patch
from tempfile import TemporaryDirectory
from ocw_data_parser.ocw_data_parser import CustomHTMLParser, OCWParser
import ocw_data_parser.test_constants as constants
import logging

log = logging.getLogger(__name__)

"""
Tests for ocw_data_parser
"""


def test_no_params(ocw_parser):
    """
    Test that an OCWParser with no params raises an exception
    """
    with pytest.raises(Exception):
        OCWParser()


def test_parser_loaded_jsons(ocw_parser):
    """
    Test instantiating a parser with preloaded jsons
    """
    assert OCWParser(
        loaded_jsons=ocw_parser.jsons), "instantiating parser with preloaded jsons failed"


def test_parser_invalid_file(ocw_parser):
    """
    Test instantiating a parser with an improperly named json file in the source directory
    """
    with TemporaryDirectory() as destination_dir:
        with open(os.path.join(constants.SINGLE_COURSE_DIR, "jsons/test.json"), "w") as f:
            with pytest.raises(ValueError):
                OCWParser(course_dir=constants.SINGLE_COURSE_DIR,
                          destination_dir=destination_dir,
                          static_prefix=constants.STATIC_PREFIX)
            os.remove(os.path.join(
                constants.SINGLE_COURSE_DIR, "jsons/test.json"))


def test_generate_master_json_none_source(ocw_parser):
    """
    Make sure that running generate_master_json doesn't throw an error after nulling
    out the parser's source jsons
    """
    ocw_parser.jsons = None
    ocw_parser.generate_master_json()
    assert ocw_parser.jsons is not None


def test_generate_master_json_none_course_image_uid(ocw_parser):
    """
    Make sure course_image_uid is regenerated by generate_master_json after it's nulled out
    """
    ocw_parser.course_image_uid = None
    ocw_parser.generate_master_json()
    assert ocw_parser.course_image_uid is not None


def test_load_raw_jsons_invalid_file(ocw_parser):
    """
    Add a json file with invalid content to the course_dir and make sure it generates an error
    """
    with TemporaryDirectory() as destination_dir:
        with open(os.path.join(constants.SINGLE_COURSE_DIR, "jsons/999.json"), "w") as f:
            f.write("{")
        with pytest.raises(json.decoder.JSONDecodeError):
            OCWParser(course_dir=constants.SINGLE_COURSE_DIR,
                      destination_dir=destination_dir,
                      static_prefix=constants.STATIC_PREFIX)
        os.remove(os.path.join(constants.SINGLE_COURSE_DIR, "jsons/999.json"))


def test_upload_all_data_to_s3(ocw_parser_s3, s3_bucket):
    """
    Use moto (mock boto) to test s3 uploading
    """
    ocw_parser_s3.upload_all_media_to_s3(upload_master_json=True)
    master_json = ocw_parser_s3.get_master_json()

    for p in master_json["course_pages"]:
        if p["text"]:
            for bucket_item in s3_bucket.objects.filter(Prefix=ocw_parser_s3.s3_target_folder):
                if bucket_item.key in p["file_location"]:
                    assert bucket_item.key == ocw_parser_s3.s3_target_folder + \
                           p["uid"] + "_" + p["short_url"] + ".html"
    for f in master_json["course_files"]:
        for bucket_item in s3_bucket.objects.filter(Prefix=ocw_parser_s3.s3_target_folder):
            if bucket_item.key in f["file_location"]:
                assert bucket_item.key == ocw_parser_s3.s3_target_folder + \
                       f["uid"] + "_" + f["id"]

        if f["uid"] == ocw_parser_s3.course_image_uid:
            assert master_json["image_src"] == s3_upload_base(
            ) + f["uid"] + "_" + f["id"]
            assert master_json["image_description"] == f["description"]
        elif f["uid"] == ocw_parser_s3.course_thumbnail_image_uid:
            assert master_json["thumbnail_image_src"] == s3_upload_base(
            ) + f["uid"] + "_" + f["id"]
            assert master_json["thumbnail_image_description"] == f["description"]

    for bucket_item in s3_bucket.objects.filter(Prefix=ocw_parser_s3.s3_target_folder):
        if bucket_item.key == ocw_parser_s3.s3_target_folder + master_json["uid"] + "_master.json":
            master_json_key = bucket_item.key
    assert master_json_key is not None

    assert master_json["image_src"] is not None
    assert master_json["thumbnail_image_src"] is not None


def test_upload_all_data_to_s3_no_binary_data(ocw_parser_s3, caplog):
    """
    Test that there is a descriptive error message when there is no binary data in a json file
    """
    with patch("ocw_data_parser.ocw_data_parser.get_binary_data", return_value=None):
        ocw_parser_s3.upload_all_media_to_s3()
        assert "Could not load binary data for file 9dbd5e22e2379a1bb4e844757c445dfd_7UJ4CFRGd-U.srt "\
        "in json file 15.json for course 18-06-linear-algebra-spring-2010" in [rec.message for rec in caplog.records]


def test_upload_all_data_to_s3_large_media_link_upload_error(ocw_parser_s3, caplog):
    """
    Test that there is a descriptive error message when a large media file cannot be uploaded
    """

    with patch("ocw_data_parser.ocw_data_parser.get", return_value=None):
        ocw_parser_s3.upload_all_media_to_s3()
        assert "Could NOT upload powerMethod.html for course 18-06-linear-algebra-spring-2010" in [
            rec.message for rec in caplog.records]



def test_upload_master_json_to_s3_no_uid(ocw_parser_s3, caplog):
    """
    Test that there is a descriptive error message when the master json has no uid

    """
    ocw_parser_s3.master_json['uid']=None
    ocw_parser_s3.upload_master_json_to_s3("testing")
    assert "No unique uid found for master_json for course 18-06-linear-algebra-spring-2010" in [
        rec.message for rec in caplog.records]


def test_upload_course_image(ocw_parser_s3, s3_bucket):
    """
    Use moto (mock boto) to test s3 uploading
    """
    ocw_parser_s3.upload_course_image()
    course_image_key = None
    master_json = ocw_parser_s3.get_master_json()

    for bucket_item in s3_bucket.objects.filter(Prefix=ocw_parser_s3.s3_target_folder):
        if bucket_item.key in ocw_parser_s3.course_image_s3_link:
            course_image_key = bucket_item.key
        elif bucket_item.key == ocw_parser_s3.s3_target_folder + master_json["uid"] + "_master.json":
            master_json_key = bucket_item.key

    assert course_image_key is not None
    assert master_json_key is not None

    assert ocw_parser_s3.master_json["image_src"] is not None
    assert ocw_parser_s3.master_json["thumbnail_image_src"] is not None

    for f in master_json["course_files"]:
        if f["uid"] == ocw_parser_s3.course_image_uid:
            assert master_json["image_src"] == s3_upload_base(
            ) + f["uid"] + "_" + f["id"]
            assert master_json["image_description"] == f["description"]
        elif f["uid"] == ocw_parser_s3.course_thumbnail_image_uid:
            assert master_json["thumbnail_image_src"] == s3_upload_base(
            ) + f["uid"] + "_" + f["id"]
            assert master_json["thumbnail_image_description"] == f["description"]
        else:
            assert f["file_location"]


def test_upload_course_image_no_s3_bucket_name(ocw_parser_s3, caplog):
    """
    Test that uploading the course image without the s3 bucket name throws an error
    """
    ocw_parser_s3.s3_bucket_name = None
    with pytest.raises(ValueError):
        ocw_parser_s3.upload_course_image()
        assert ["Please set your s3 bucket name"] == [
            rec.message for rec in caplog.records]


def test_get_master_json(ocw_parser):
    """
    Test that getting the master JSON is not None or empty or missing basic properties
    and doesn't throw an exception
    """
    fail_template = "failed to read {} property from master json"
    master_json = ocw_parser.get_master_json()
    assert master_json, "failed to get master json"
    assert master_json["uid"], fail_template.format("uid")
    assert master_json["title"], fail_template.format("title")
    assert master_json["description"], fail_template.format("description")
    assert master_json["short_url"], fail_template.format("short_url")
    assert len(master_json["open_learning_library_related"]) == 3, fail_template.format("open_learning_library_related")
    assert master_json["open_learning_library_related"][0]["url"] == "https://openlearninglibrary.mit.edu/courses/course-v1:MITx+18.01.1x+2T2019/about"

def test_export_master_json_s3_links(ocw_parser_s3):
    """
    Test that exporting the master json file with s3 links doesn't error
    """
    ocw_parser_s3.export_master_json(s3_links=True)
    assert os.path.isdir(os.path.join(ocw_parser_s3.destination_dir, "master"))


def test_export_master_json_no_s3_bucket_name(ocw_parser_s3, caplog):
    """
    Test that exporting the master json file without an s3 bucket name throws an error
    """
    ocw_parser_s3.s3_bucket_name = None
    ocw_parser_s3.export_master_json(s3_links=True)
    assert ["Please set your s3 bucket name"] == [
        rec.message for rec in caplog.records]


def test_set_s3_bucket_name(ocw_parser_s3):
    """
    Test setting the s3 bucket name
    """
    assert "testing" == ocw_parser_s3.s3_bucket_name


def test_set_s3_access_key(ocw_parser_s3):
    """
    Test setting the s3 access key
    """
    assert "testing" == ocw_parser_s3.s3_bucket_access_key


def test_set_s3_secret_access_key(ocw_parser_s3):
    """
    Test setting the s3 secret access key
    """
    assert "testing" == ocw_parser_s3.s3_bucket_secret_access_key


def test_set_s3_target_folder(ocw_parser_s3):
    """
    Test setting the s3 target folder
    """
    assert "course-1" == ocw_parser_s3.s3_target_folder


def s3_upload_base():
    return "https://testing.s3.amazonaws.com/course-1/"


def test_uid(ocw_parser, course_id):
    """
    Test that the uid property of the master JSON matches the uid of the course site root
    """
    ocw_parser.export_master_json()
    with open(os.path.join(constants.SINGLE_COURSE_DIR, "jsons/1.json"), "r") as first_json:
        first_json_data = json.loads(first_json.read())
        with open(os.path.join(ocw_parser.destination_dir, "master/master.json"), "r") as master_json:
            master_json_data = json.loads(master_json.read())
            assert first_json_data["_uid"] == master_json_data["uid"]


def test_course_files(ocw_parser):
    """Make sure course_files include the right fields with the correct default values"""
    assert len(ocw_parser.master_json['course_files']) == 172
    assert ocw_parser.master_json['course_files'][0] == {
        'order_index': 6,
        'uid': 'c24518ecda658185c40c2e5eeb72c7fa',
        'id': '182.png',
        'parent_uid': '3f3b7835cf477d3ba10b05fbe03cbffa',
        'title': '182.png',
        'caption': '',
        'file_type': 'image/png',
        'alt_text': '',
        'credit': '',
        'platform_requirements': '',
        'description': '',
        'type': 'OCWImage',
    }
    
def test_course_foreign_files(ocw_parser):
    """Make sure course_foreign_files include the right fields with the correct default values"""
    assert ocw_parser.master_json['course_foreign_files'][0] == {
        'link': 'http://ocw.mit.edu/ans7870/18/18.06/tools/Applets_sound/uropmovie.html',
        'parent_uid': 'b5785e071ddb991cf3dfd7cc469e6397'
    }
    assert len(ocw_parser.master_json['course_foreign_files']) == 20


def test_other_information_text(ocw_parser):
    """other_information_text should be an empty string"""
    assert ocw_parser.master_json['other_information_text'] == ''


def test_course_pages(ocw_parser):
    """assert the output of composing course_pages"""
    assert len(ocw_parser.master_json["course_pages"])
    page = ocw_parser.master_json["course_pages"][0]
    page_without_text = {**page}
    del page_without_text["text"]
    del page_without_text["description"]
    assert page_without_text == {
        'order_index': 3,
        'uid': 'ede17211bd49ea166ed701f09c1de288',
        'parent_uid': 'aabc44bdb2e45374d62f30f2a6d4c63e',
        'title': 'Syllabus',
        'short_page_title': 'Syllabus',
        'type': 'CourseSection',
        'is_image_gallery': False,
        'is_media_gallery': False,
        'list_in_left_nav': False,
        'file_location': 'ede17211bd49ea166ed701f09c1de288_syllabus.html',
        'short_url': 'syllabus',
        'url': '/courses/mathematics/18-06-linear-algebra-spring-2010/syllabus',
        'bottomtext': '<p>Sample Bottom Text</p>'
    }
    assert page["text"].startswith("<h2 class=\"subhead\">Course Meeting Times")
    assert page["description"].startswith("This syllabus section provides information on course goals")


@pytest.mark.parametrize("has_instructors", [True, False])
def test_instructors(ocw_parser, has_instructors):
    """instructors list should be present as a list in the output"""
    expected_instructor = {**ocw_parser.jsons[0]["instructors"][0]}
    if not has_instructors:
        ocw_parser.jsons[0]["instructors"] = []
    ocw_parser.generate_master_json()
    del expected_instructor["mit_id"]
    assert ocw_parser.master_json["instructors"] == ([expected_instructor] if has_instructors else [])


def test_course_features(ocw_parser):
    """assert the output of course_features"""
    assert ocw_parser.master_json["course_features"] == [
        {
            "ocw_feature": "AV special element video",
            "ocw_subfeature": "Tutorial",
            "ocw_feature_url": "./resolveuid/b5785e071ddb991cf3dfd7cc469e6397",
            "ocw_speciality": "",
            "ocw_feature_notes": ""
        },
        {
            "ocw_feature": "AV lectures",
            "ocw_subfeature": "Video",
            "ocw_feature_url": "./resolveuid/6b1f662457366951bfe85945521b0299",
            "ocw_speciality": "",
            "ocw_feature_notes": ""
        },
        {
            "ocw_feature": "Assignments",
            "ocw_subfeature": "problem sets with solutions",
            "ocw_feature_url": "./resolveuid/87609dbba9d13a6b234d62de21a20433",
            "ocw_speciality": "",
            "ocw_feature_notes": ""
        },
        {
            "ocw_feature": "Exams",
            "ocw_subfeature": "Solutions",
            "ocw_feature_url": "./resolveuid/c13c4766c0cf1486f0cf6435c531eaad",
            "ocw_speciality": "",
            "ocw_feature_notes": ""
        },
        {
            "ocw_feature": "Instructor Insights",
            "ocw_subfeature": "",
            "ocw_feature_url": "./resolveuid/3f3b7835cf477d3ba10b05fbe03cbffa",
            "ocw_speciality": "",
            "ocw_feature_notes": ""
        }
    ]


def test_tags(ocw_parser):
    """assert tags output"""
    expected_tags = [
        "matrix theory",
        "linear algebra",
        "systems of equations",
        "vector spaces",
        "determinants",
        "eigenvalues",
        "similarity",
        "positive definite matrices",
        "least-squares approximations",
        "stability of differential equations",
        "networks",
        "Fourier transforms",
        "Markov processes",
    ]
    assert ocw_parser.master_json["tags"] == [{"name": tag} for tag in expected_tags]


def test_course_embedded_media(ocw_parser):
    """assert embedded media"""
    assert len(ocw_parser.master_json["course_embedded_media"]) == 36
    key = "12700054aninterviewwithgilbertstrangonteachinglinearalgebra63021644"
    media_json = {**ocw_parser.master_json["course_embedded_media"][key]}
    transcript = media_json["transcript"]
    del media_json["transcript"]
    embedded_media = media_json["embedded_media"]
    del media_json["embedded_media"]
    assert media_json == {
        'about_this_resource_text': '',
        'inline_embed_id': '12700054aninterviewwithgilbertstrangonteachinglinearalgebra63021644',
        'order_index': 10,
        'parent_uid': '3f3b7835cf477d3ba10b05fbe03cbffa',
        'related_resources_text': '',
        'short_url': 'an-interview-with-gilbert-strang-on-teaching-linear-algebra',
        'technical_location': 'https://ocw.mit.edu/courses/mathematics/18-06-linear-algebra-spring-2010/instructor-insights/an-interview-with-gilbert-strang-on-teaching-linear-algebra',
        'title': 'An Interview with Gilbert Strang on Teaching Linear Algebra',
        'uid': 'e21b71ff0fa975bfa9acb2a155aafc1d',
    }
    assert transcript.startswith("<p><span m=\'6840\'>SARAH HANSEN:")
    assert len(embedded_media) == 11
    assert embedded_media[1] == {  # important because it has a technical_location
        'id': '18.06.jpg',
        'parent_uid': 'e21b71ff0fa975bfa9acb2a155aafc1d',
        'technical_location': 'https://ocw.mit.edu/courses/mathematics/18-06-linear-algebra-spring-2010/instructor-insights/an-interview-with-gilbert-strang-on-teaching-linear-algebra/18.06.jpg',
        'title': '18.06.jpg',
        'type': None,
        'uid': 'f777380de6feec2c42ab6e159e05ddf2'
    }
    assert embedded_media[2] == {  # important because it has a media_location
        'id': 'Thumbnail-YouTube-JPG',
        'media_location': 'https://img.youtube.com/vi/7UJ4CFRGd-U/default.jpg',
        'parent_uid': 'e21b71ff0fa975bfa9acb2a155aafc1d',
        'title': 'Thumbnail-YouTube-JPG',
        'type': 'Thumbnail',
        'uid': '7cd0685535147aebd9d7c2e98dc68afd'
    }


def test_foreign_files(ocw_parser):
    """assert course_foreign_files output"""
    assert len(ocw_parser.master_json["course_foreign_files"]) == 20
    assert ocw_parser.master_json["course_foreign_files"][0] == {
        'link': 'http://ocw.mit.edu/ans7870/18/18.06/tools/Applets_sound/uropmovie.html',
        'parent_uid': 'b5785e071ddb991cf3dfd7cc469e6397',
    }


def test_extract_media_locally(ocw_parser):
    """extract_media_locally should write media files to a local directory"""
    ocw_parser.extract_media_locally()
    static_files = Path(ocw_parser.destination_dir) / "output" / "static_files"
    for path in static_files.iterdir():
        assert path.stat().st_size > 0  # make sure files are non-trivial

    expected_counts = {
        ".pdf": 93,
        ".srt": 36,
        ".html": 11,
        ".jpg": 42,
        ".png": 1,
    }
    counts = {}
    for path in static_files.iterdir():
        ext = os.path.splitext(path)[1]
        if ext not in counts:
            counts[ext] = 0
        counts[ext] += 1
    assert counts == expected_counts
